//==============================================================
//  $Id: lcd.c,v 1.1 2008/04/27 13:44:12 noritan Exp $
//  $Source: /mnt/cvs/rep/CW/HID08/Sources/lcd.c,v $
//==============================================================
#include <hidef.h> /* for EnableInterrupts macro */
#include "derivative.h" /* include peripheral declarations */

//--------------------------------------------------------------
//  Time constants
//--------------------------------------------------------------
#define   _CPU_CLOCK_   (12000000UL)
#define   _BUS_CLOCK_   (6000000UL)
#define   PERIOD_PS     (1)
#define   PERIOD_1m00   ((_BUS_CLOCK_ / PERIOD_PS / 1000) - 1)

//--------------------------------------------------------------
//  Alias for I/O ports.
//--------------------------------------------------------------
#define   LCD_RS        PTE_PTE0
#define   LCD_RW        PTE_PTE2
#define   LCD_E         PTC_PTC0
#define   LCD_DB        PTA
#define   LCD_DDRRS     DDRE_DDRE0
#define   LCD_DDRRW     DDRE_DDRE2
#define   LCD_DDRE      DDRC_DDRC0
#define   LCD_DDRDB     DDRA

//--------------------------------------------------------------
//  MSEC sleep.
//--------------------------------------------------------------
void msleep(dword reltim) {
  while (reltim > 0) {
    while (T1SC_TOF) ;
    T1SC_TOF = 0;
    reltim--;
  }
}

//--------------------------------------------------------------
//  Toggle E clock.
//--------------------------------------------------------------
static void lcd_toggleE(void) {
  LCD_E = 1;                    // assert LCD enable
  LCD_E = 0;                    // negate LCD enable
}

//--------------------------------------------------------------
//  Get a byte.
//--------------------------------------------------------------
byte lcd_getByte(void) {
  byte data;
  LCD_E = 1;                    // assert LCD enable
  data = LCD_DB;              // Get Status.
  LCD_E = 0;                    // negate LCD enable
  return data;
}

//--------------------------------------------------------------
//  Get two nibbles.
//--------------------------------------------------------------
byte lcd_getNibbles(void) {
  byte data;
  LCD_E = 1;                    // assert LCD enable
  data = LCD_DB & 0xF0;         // Get MSB part.
  LCD_E = 0;                    // negate LCD enable
  LCD_E = 1;                    // assert LCD enable
  data |= (LCD_DB >> 4);        // Get LSB part.
  LCD_E = 0;                    // negate LCD enable
  return data;
}

//--------------------------------------------------------------
//  Get status word.
//--------------------------------------------------------------
byte lcd_getStatus(void) {
  byte status;
  
  LCD_RS = 0;                   // Select command.
  status = lcd_getNibbles();    // Get Status as nibbles.
  return status;
}

//--------------------------------------------------------------
//  Get data word.
//--------------------------------------------------------------
byte lcd_getData(void) {
  byte data;
  
  LCD_RS = 1;                   // Select register.
  data = lcd_getNibbles();      // Get Data as nibbles.
  return data;
}

//--------------------------------------------------------------
//  Check if LCD is busy.
//--------------------------------------------------------------
byte lcd_isBusy(void) {
  return (lcd_getStatus() & 0x80);
}

//--------------------------------------------------------------
//  Send two nibbles
//--------------------------------------------------------------
void lcd_sendNibbles(byte data) {
  // Set MSB nibble
  LCD_DB = (LCD_DB & 0x0F) | (data & 0xF0);
  LCD_DDRDB |= 0xF0;            // Drive DB(7:4) as output
  lcd_toggleE();                // Toggle LCD E clock.
  // Set LSB nibble
  LCD_DB = (LCD_DB & 0x0F) | ((data << 4) & 0xF0);
  lcd_toggleE();                // Toggle LCD E clock.
  LCD_DDRDB &= 0x0F;            // Release DB(7:4) as input
}

//--------------------------------------------------------------
//  Send one bytes
//--------------------------------------------------------------
void lcd_sendByte(byte data) {
  LCD_DB = data;                // Set a byte.
  LCD_DDRDB = 0xFF;             // Drive DB as output
  lcd_toggleE();                // Toggle LCD E clock.
  LCD_DDRDB = 0x00;             // Release DB as input
}

//--------------------------------------------------------------
//  Send a byte as a command
//--------------------------------------------------------------
void lcd_sendCommand(byte command) {
  LCD_RS = 0;                   // Select command.
  LCD_RW = 0;                   // Prepare for WRITE.
  lcd_sendNibbles(command);     // Send as nibbles.
  LCD_RW = 1;                   // Prepare for READ.
}

//--------------------------------------------------------------
//  Send a byte as a data
//--------------------------------------------------------------
void lcd_sendData(byte data) {
  LCD_RS = 1;                   // Select data.
  LCD_RW = 0;                   // Prepare for WRITE.
  lcd_sendNibbles(data);        // Send as nibbles.
  LCD_RW = 1;                   // Prepare for READ.
}

//--------------------------------------------------------------
//  Locate cursor 
//--------------------------------------------------------------
void lcd_cursor(byte addr) {
  while (lcd_isBusy()) ;        // Wait for instruction ends.
  lcd_sendCommand(addr | 0x80); // Send an address set command.
}

//--------------------------------------------------------------
//  Initialization sequence
//--------------------------------------------------------------
void lcd_init(void) {
  void putc(char);

  LCD_E = 0;                    // clear LCD enable
  LCD_RS = 0;                   // clear register select.
  LCD_RW = 1;                   // Prepare for READ.
  LCD_DDRE = 1;                 // configure LCD_E as output.
  LCD_DDRRS = 1;                // configure LCD_RS as output.
  LCD_DDRRW = 1;                // configure LCD_RW as output.
  
  PTD_PTD5 = 1;
  PTD_PTD4 = 1;
  PTD_PTD3 = 1;
  PTD_PTD2 = 1;
  DDRD_DDRD5 = 1;               // Enable BLUE
  DDRD_DDRD4 = 1;               // Enable GREEN
  DDRD_DDRD3 = 1;               // Enable YELLOW
  DDRD_DDRD2 = 1;               // Enable RED
  
  // Initialization sequence.
  LCD_RS = 0;                   // Select command.
  LCD_RW = 0;                   // Prepare for WRITE.
  msleep(20);                   // wait for 20ms
  LCD_DB = (LCD_DB & 0x0F) | (0x30);  // 1st command.
  LCD_DDRDB |= 0xF0;            // Drive DB(7:4) as output
  lcd_toggleE();                // Toggle LCD E clock.
  PTD_PTD4 = 0;
  msleep(10);                   // wait for 10ms
  lcd_toggleE();                // Toggle LCD E clock.
  msleep(1);                    // wait for 1ms
  LCD_DB = (LCD_DB & 0x0F) | (0x20);  // 4-bit mode.
  lcd_toggleE();                // Toggle LCD E clock.
  LCD_DDRDB &= 0x0F;            // Release DB(7:4) as input
  LCD_RW = 1;                   // Prepare for READ.
  msleep(1);                    // wait for 1ms

  while (lcd_isBusy()) ;        // Wait for instruction ends.
  lcd_sendCommand(0x28);        // Set DL as 4-bit, N as 2 lines, F as 5x7 dots.
  
  while (lcd_isBusy()) ;        // Wait for instruction ends.
  PTD_PTD3 = 0;
  lcd_sendCommand(0x10);        // Set S/C as move, R/L as left.
  
  while (lcd_isBusy()) ;        // Wait for instruction ends.
  PTD_PTD2 = 0;
  lcd_sendCommand(0x0F);        // Set D as ON, C as ON, B as blink.
  
  while (lcd_isBusy()) ;        // Wait for instruction ends.
  lcd_sendCommand(0x06);        // Set I/D as INC, S as no Shift.

  while (lcd_isBusy()) ;        // Wait for instruction ends.
  lcd_sendCommand(0x01);        // Clear Display.
}

static int pos = 0;

void lcd_putc(byte ch) {
  while (lcd_isBusy());
  lcd_sendData(ch);
  pos++;
  if (pos == 16) {
    lcd_cursor(0x40);
  } else if (pos >= 32) {
    lcd_cursor(0x00);
    pos = 0;
  }
}