//
// $Source: /mnt/cvs/rep/CW/HID08/Sources/usb_lowlevel.c,v $
// $Id: usb_lowlevel.c,v 1.1 2008/04/27 13:44:12 noritan Exp $
//
// Low level function for USB
//
#include <hidef.h>
#include "derivative.h"
#include "usb_lowlevel.h"
#include "usb_descriptors.h"                                 
#include "lcd.h"

byte usbState;
#pragma DATA_SEG __SHORT_SEG MY_ZEROPAGE

enum ControlState {
  ST_IDLE = 0,
  ST_NODATA_STATUS,
  ST_READ_DATA,
  ST_READ_STATUS,
  ST_WRITE_DATA,
  ST_WRITE_STATUS
};
volatile enum ControlState controlState;

volatile byte suspendCounter;

SetupBuffer setupBuffer;
struct Flag {
    byte    setupReady:1;
    byte    ep0OutReady:1;
    byte    ep0InComplete:1;
    byte    flag_dummy:5;
}       flag;

const byte *ep0InXferPointer;
word ep0InXferCount;
byte ep0InXferZLP;

byte (*nodataStatusProc)(void);
byte (*writeDataProc)(void);
void (*writeStatusProc)(void);

#pragma DATA_SEG DEFAULT

void idleInit(void);
void nodataStatusInit(byte (*proc)(void));
void readDataInit(const byte *buf, word len);
void readStatusInit(void);
void writeDataInit(byte (*dataProc)(void), void (*statusProc)(void));
void writeStatusInit(void);

//----------------------------------------------------------------------------
//  Error code
//----------------------------------------------------------------------------
#define     ERR_NONE        (0)
#define     ERR_STALL       (1)


//----------------------------------------------------------------------------
//  Copy SETUP packet into buffer.
//  Intel words are converted into MOTOROLA words.
//----------------------------------------------------------------------------
void copySetupBuffer(void) {
  setupBuffer.bmRequestType     = UE0D0;
  setupBuffer.bRequest          = UE0D1;
  setupBuffer.wValue            = (UE0D3 << 8) | UE0D2;
  setupBuffer.wIndex            = (UE0D5 << 8) | UE0D4;
  setupBuffer.wLength           = (UE0D7 << 8) | UE0D6;
}


//----------------------------------------------------------------------------
//  put EP0 value into a buffer
//----------------------------------------------------------------------------
void putEp0OutBuffer(void) {
  switch (USR0_RP0SIZ) {
    case 8:
      ep0OutBuffer[7] = UE0D7;
    case 7:
      ep0OutBuffer[6] = UE0D6;
    case 6:
      ep0OutBuffer[5] = UE0D5;
    case 5:
      ep0OutBuffer[4] = UE0D4;
    case 4:
      ep0OutBuffer[3] = UE0D3;
    case 3:
      ep0OutBuffer[2] = UE0D2;
    case 2:
      ep0OutBuffer[1] = UE0D1;
    case 1:
      ep0OutBuffer[0] = UE0D0;
  }
}



//----------------------------------------------------------------------------
//  Force STALL Condition for EP0 (both IN and OUT)
//  as a Response to an invalid SETUP Request
//  Flags will be auto-cleared by the next SETUP Token
//----------------------------------------------------------------------------
void forceStall(void) {
  UCR3_OSTALL0 = 1;
  UCR3_ISTALL0 = 1;
  PTD_PTD2 = 0;
}

//----------------------------------------------------------------------------


// CLEAR_FEATURE Standard Device Request Handler
// called by handleSETUP();
//
void clearFeature() {

	if( setupBuffer.wValue >= 256 ||
		setupBuffer.wIndex >= 256 ||
		setupBuffer.wLength != 0)	// check 0-fields
		forceStall();
	else if((setupBuffer.bmRequestType == RT_ENDPOINT) &&
		(setupBuffer.wValue == ENDPOINT_HALT) &&
		((setupBuffer.wIndex == 0x81) || (setupBuffer.wIndex == 0x02))) {

		UCR1_T1SEQ=0;	// clear STALL, Sequence = DATA0
		UCR1_STALL1=0;
		UCR2_T2SEQ=0;	// clear STALL, Sequence = DATA0
		UCR2_STALL2=0;
		// prepare to send empty DATA1 at next IN Transaction (ACK)
		UCR0=0;
		UCR0_T0SEQ=1;
		UCR0_TX0E=1;
		}
	else forceStall();
	}


byte illegalTransaction(void) {
  return ERR_STALL;             // Protocol error
}



//----------------------------------------------------------------------------
//  SET_ADDRESS Standard Device Request Handler
//  called by SETUP handler.
//  according to USB2.0 spec page 256
//----------------------------------------------------------------------------
byte setAddressSetup(void) {
  byte setAddressStatus(void);
  
  // SET_ADDRESS transaction validation
  if (setupBuffer.wIndex != 0)      return ERR_STALL;
  if (setupBuffer.wLength != 0)     return ERR_STALL;
  if (setupBuffer.wValue >= 128)    return ERR_STALL;
  
  // Confirm STATE
  switch (usbState) {
    case US_DEFAULT:
    case US_ADDRESSED:
      break;
    default:                        return ERR_STALL;
  }
  
lcd_putc('s');

  // State transition.
  nodataStatusInit(setAddressStatus);
    
  // Successfully processed.
  return ERR_NONE;
}

//----------------------------------------------------------------------------
//  SET_ADDRESS Standard Device Request Handler
//  called by STATUS handler.
//  according to USB2.0 spec page 256
//----------------------------------------------------------------------------
byte setAddressStatus(void) {
    // A STATUS transaction was received.
    // Set the USB address.
    UADDR = (byte)setupBuffer.wValue | UADDR_USBEN_MASK;
    if (setupBuffer.wValue == 0) {
      // make a transition to DEFAULT
      usbState = US_DEFAULT;
    } else {
      // make a transition to ADDRESSED
      usbState = US_ADDRESSED;
    }
    // Reset the request code.
    setupBuffer.bRequest = REQUEST_COMPLETE;
    
lcd_putc('A');

    // State transition.
    idleInit();
    
    return ERR_NONE;
}

//----------------------------------------------------------------------------
//  SET_CONFIGURATION Standard Device Request Handler
//  called by SETUP handler.
//  according to USB2.0 spec page 257
//----------------------------------------------------------------------------
byte setConfigurationSetup(void) {
  byte setConfigurationStatus(void);
  
  // SET_CONFIGURATION transaction validation
  if (setupBuffer.wIndex != 0)      return ERR_STALL;
  if (setupBuffer.wLength != 0)     return ERR_STALL;
  if (setupBuffer.wValue >= 256)    return ERR_STALL;

  // Validate STATE
  switch (usbState) {
    case US_ADDRESSED:
    case US_CONFIGURED:
      break;
    default:                        return ERR_STALL;
  }

  // Validate configuration number
  switch (setupBuffer.wValue) {
    case 0:
    case 1:
      break;
    default:                        return ERR_STALL;
  }
  
lcd_putc('s');

  nodataStatusInit(setConfigurationStatus);
  
  return ERR_NONE;
}

//----------------------------------------------------------------------------
//  SET_CONFIGURATION Standard Device Request Handler
//  called by STATUS handler.
//  according to USB2.0 spec page 257
//----------------------------------------------------------------------------
byte setConfigurationStatus(void) {
  switch (setupBuffer.wValue) {
    case 0:
      UCR3_ENABLE1 = 0;         // Disable EP1
      UCR3_ENABLE2 = 0;         // Disable EP2
      usbState = US_ADDRESSED;
      break;
    case 1:
      UCR3_ENABLE1 = 1;         // Enable EP1
      UCR3_ENABLE2 = 1;         // Enable EP2
      usbState = US_CONFIGURED;
      break;
  }
lcd_putc('C');
  return ERR_NONE;
}

void fillEp0InBuffer(void) {
  byte size;
  
  UCR0_TX0E = 0;                // Inactivate EP0-IN
  UIR2_TXD0FR = 1;              // Acknowledge
  if (ep0InXferCount == 0) {    // all data sent.
    if (ep0InXferZLP) {         // ZLP requested ?
      // Send ZLP
      ep0InXferZLP = FALSE;     // This is the last packet.
      // Activate EP0-IN
      UCR0_T0SEQ = ~UCR0_T0SEQ; // Set sequence bit.
      UCR0_TP0SIZ = 0;          // Set ZERO to size
      UCR0_TX0E = 1;            // Activate EP0-IN
      return;
    } else {
      // all packet completed
      flag.ep0InComplete = 1;   // Assert EP0-IN completion flag
      return;
    }
  } else {    
    // copy (up to) 8 Bytes to EP0 Data Registers
    size = 0;
    while ((ep0InXferCount > 0) && (size < 8)) {
      (&UE0D0)[size] = *ep0InXferPointer++;
      ep0InXferCount--;
      size++;
    }
    // Activate EP0-IN
    UCR0_T0SEQ = ~UCR0_T0SEQ;   // Set sequence bit.
    UCR0_TP0SIZ = size;         // Set packet size.
    UCR0_TX0E = 1;              // Activate EP0-IN
    return;
  }
}

//----------------------------------------------------------------------------
//  GET_DESCRIPTOR Standard Device Request Handler
//  called by SETUP handler.
//  according to USB2.0 spec page 253
//  according to HID1.11 spec page 49
//----------------------------------------------------------------------------
byte getDescriptorSetup(void) {
  byte descriptorIndex = setupBuffer.wValue & 0xFF;
  word interfaceNumber = setupBuffer.wIndex;
  
  // Validate STATE
  switch (usbState) {
    case US_DEFAULT:
    case US_ADDRESSED:
    case US_CONFIGURED:
      break;
    default:                        return ERR_STALL;
  }

lcd_putc('g');
lcd_putc('D');
  switch (setupBuffer.bmRequestType & 0x0F) {
    case 0:     // recipient : device
      // Validate and select a descriptor
      // Language code (wIndex) must be ZERO except DT_STRING
      // Descriptor index (wValue[7:0]) must be ZERO for DT_DEVICE
      switch (setupBuffer.wValue >> 8) {
        // Standard Descriptors
        case DT_DEVICE:
          if (setupBuffer.wIndex != 0)  return ERR_STALL;
          if (descriptorIndex != 0)     return ERR_STALL;
lcd_putc('D');
          readDataInit((byte *)&deviceDesc, deviceDesc.bLength);
    			break;
        case DT_CONFIGURATION:
          if (setupBuffer.wIndex != 0)  return ERR_STALL;
lcd_putc('C');
          switch (descriptorIndex) {
            case 0:
              readDataInit(
                (byte *)&configDesc,
                configDesc.wTotalLength.lo
              );
        			break;
        		default:                    return ERR_STALL;
          }
          break;
        case DT_STRING:
lcd_putc('S');
          if (descriptorIndex >= strDescTableSize) {
                                        return ERR_STALL;
          }
          readDataInit(
            strDescTable[descriptorIndex],
            *(strDescTable[descriptorIndex])
          );
    			break;
        default:                        return ERR_STALL;
        
      }
      break;
    case 1:     // recipient : interface
      // Validate and select a descriptor
      // Interface number (wIndex) indicates the interface #.
      // Descriptor index (wValue[7:0]) must be ZERO for DT_HID
      switch (setupBuffer.wValue >> 8) {
        // Standard Descriptors
        case DT_HID:
lcd_putc('H');
          if (descriptorIndex != 0)     return ERR_STALL;
          switch (interfaceNumber) {
            case 0:     // Interface #0 selected.
              readDataInit((byte *)&hidDesc0, sizeof(hidDesc0));
              break;
            case 1:     // Interface #1 selected.
              readDataInit((byte *)&hidDesc1, sizeof(hidDesc1));
              break;
            default:    // Illegal interface index.
              return ERR_STALL;
          }
    			break;
        case DT_REPORT:
lcd_putc('R');
          if (descriptorIndex != 0)     return ERR_STALL;
          switch (interfaceNumber) {
            case 0:     // Interface #0 selected.
              readDataInit((byte *)&reportDesc0, sizeof(reportDesc0));
              break;
            case 1:     // Interface #1 selected.
              readDataInit((byte *)&reportDesc1, sizeof(reportDesc1));
              break;
            default:    // Illegal interface index.
              return ERR_STALL;
          }
          break;
        case DT_PHYSICAL:
        default:                        return ERR_STALL;
        
      }
      break;
    default:                            return ERR_STALL;
  }

  // completed.
  return ERR_NONE;
}


//----------------------------------------------------------------------------
//  Make a transition to powered but not reset yet.
//----------------------------------------------------------------------------
void initUsb() {
  // set default address
  UADDR         = 0;

  // disable Interrupts
  UIR0  = 0b00000000;
  //        |||||||+-- RXD0IE   = 0 disable RXD0 interrupt
  //        ||||||+--- TXD0IE   = 0 disable TXD0 interrupt
  //        |||||+---- Reserved (0)
  //        ||||+----- TXD1IE   = 0 disable TXD1 interrupt
  //        |||+------ RXD2IE   = 0 disable RXD2 interrupt
  //        ||+------- TXD2IE   = 0 disable TXD2 interrupt
  //        |+-------- SUSPEND  = 0 disable Suspend function
  //        +--------- EOPIE    = 0 disable EOP interrupt

  // clear all Flags in UIR1
  UIR2  = 0b11111111;
  //        |||||||+-- RXD0FR=1     clear RXD0 flag
  //        ||||||+--- TXD0FR=1     clear TXD0 flag
  //        |||||+---- RESUMEFR=1   clear RESUME flag
  //        ||||+----- TXD1FR=1     clear TXD1 flag
  //        |||+------ RXD2FR=1     clear RXD2 flag
  //        ||+------- TXD2FR=1     clear TXD2 flag
  //        |+-------- RSTFR=1      clear bus-reset flag
  //        +--------- EOPFR=1      clear EOP flag

  // initialize endpoint #0
  UCR0  = 0b00000000;
  //        ||||++++-- TP0SIZ=0     reset packet size register
  //        |||+------ RX0E=0       disable EP0 receiver
  //        ||+------- TX0E=0       disable EP0 transmitter
  //        |+-------- Reserved (0)
  //        +--------- T0SEQ=0      reset TX sequence bit
  
  // initialize endpoint #1
  UCR1  = 0b00000000;
  //        ||||++++-- TP1SIZ=0     reset packet size register
  //        |||+------ FRESUM=0     disable Force RESUME
  //        ||+------- TX1E=0       disable EP1 transmitter
  //        |+-------- STALL1=0     disable Force EP1 STALL
  //        +--------- T1SEQ=0      reset TX sequence bit

  // initialize endpoint #2
  UCR2  = 0b00000000;
  //        ||||++++-- TP2SIZ=0     reset packet size register
  //        |||+------ RX2E=0       disable EP2 receiver
  //        ||+------- TX2E=0       disable EP2 transmitter
  //        |+-------- STALL2=0     disable Force EP2 STALL
  //        +--------- T2SEQ=0      reset TX sequence bit

  // initialize others
  UCR3  = 0b01000100;
  //        |||||||+-- ENABLE1=0    disable EP1
  //        ||||||+--- ENABLE2=0    disable EP2
  //        |||||+---- PULLEN=1     enable pull-up at D-
  //        ||||+----- Reserved (0)
  //        |||+------ ISTALL0=0    disable EP0 Force STALL for IN
  //        ||+------- OSTALL0=0    disable EP0 Force STALL for OUT
  //        |+-------- TX1STR=1     clear TX1ST flag
  //        +--------- Reserved (0)

  // initialize force bits.
  UCR4  = 0b00000000;
  //        |||||||+-- FDM=0        disable Force D- port
  //        ||||||+--- FDP=0        disable Force D+ port
  //        |||||+---- FUSBO=0      disable Force USB output
  //        +++++----- Reserved (00000)

  // make a transition	
  usbState      = US_POWERED;   // powered, but not yet reset

  UADDR_USBEN   = 1;            // enable USB interface
}


//----------------------------------------------------------------------------
//  USB bus-reset handler
//----------------------------------------------------------------------------
void resetUsb(void) {
  initUsb();                    // Soft Reset of USB Systems
  
  // INT enabled for: end of packet
  UIR0  = 0b00000011; //0x8B;
  //        |||||||+-- RXD0IE=1     enable RXD0 interrupt
  //        ||||||+--- TXD0IE=1     enable TXD0 interrupt
  //        |||||+---- Reserved (0)
  //        ||||+----- TXD1IE=0     disable TXD1 interrupt
  //        |||+------ RXD2IE=0     disable RXD2 interrupt
  //        ||+------- TXD2IE=0     disable TXD2 interrupt
  //        |+-------- SUSPEND=0    disable Suspend function
  //        +--------- EOPIE=0      disable EOP interrupt

  
  // enable EP0 receiver only.
  UCR0_RX0E     = 1;            // EP0 Receive Enable

  // make a transition
  usbState      = US_DEFAULT;   // Device is powered and reset
}


//----------------------------------------------------------------------------
//  SETUP transaction parser
//----------------------------------------------------------------------------
byte parseSetup(void) {
  switch (setupBuffer.bmRequestType & 0x60) {
    case RT_STANDARD:  // Standard Request Decoder:
      switch (setupBuffer.bRequest) {
        case SET_ADDRESS:       // 5
          return setAddressSetup();
        case GET_DESCRIPTOR:    // 6
          return getDescriptorSetup();
        case SET_CONFIGURATION: // 9
          return setConfigurationSetup();
        default:
          return ERR_STALL;
      }
      break;
    case RT_CLASS:  // Class Request Decoder:
      switch (setupBuffer.bRequest) {
        default:
          return ERR_STALL;
      }
      break;
    default:    // Unsupported Request
      return ERR_STALL;
  }
}


//----------------------------------------------------------------------------
//  Process for SETUP transaction.
//----------------------------------------------------------------------------
byte getSetup(void) {
  UIR2_RXD0FR = 1;              // Acknowledge
  if(USR0_RP0SIZ != 8) {        // SETUP Transaction must be Size=8
    return ERR_STALL;           // otherwise we have an Error Condition
  }
  copySetupBuffer();            // Copy SETUP parameters
  flag.setupReady = 1;          // Notify to main task.
  return ERR_NONE;              // Return quickly.
}


//----------------------------------------------------------------------------
//  Process for EP0-OUT transaction.
//----------------------------------------------------------------------------
byte getEp0(void) {
  UIR2_RXD0FR = 1;              // Acknowledge
  putEp0OutBuffer();            // Copy EP0-OUT payload.
  flag.ep0OutReady = 1;         // Notify to main task.
  return ERR_NONE;
}

void ep0InStart(const byte *buf, const word len, const byte zlp) {  
  ep0InXferPointer = buf;
  ep0InXferCount = len;
  ep0InXferZLP = zlp;
  fillEp0InBuffer();            // Fill first packet
}

//----------------------------------------------------------------------------
// USB Interrupt Handler
// All Interrupt Sources of the JB16's integrated USB peripheral
// will be treated by this ISR
//----------------------------------------------------------------------------
void interrupt VectorNumber_USB isrUSB() {
  for (;;) {      
    // Check RXD at EP0
    if (UIR1_RXD0F) {           // any data received via EP0?
      if (USR0_SETUP) {         // SETUP packet?
        switch (getSetup()) {   // get setup packet
          case ERR_NONE:
            break;
          default:              // STALL for illegal SETUP
            forceStall();       // force STALL
            break;
        }
      } else {                  // or a normal packet
        switch (getEp0()) {     // receive the transaction
          case ERR_NONE:
            break;
          default:              // STALL for illegal EP0-OUT
            forceStall();       // force STALL
            break;
        }
      }
      continue;                 // try other events
    }
    
    if (UIR1_TXD0F) {           // any data transmitted via EP0 ?
      fillEp0InBuffer();        // fill EP0-IN buffer.
    }

    if(UIR1_EOPF) {             // End of Packet detected?
      suspendCounter = 0;       // reset 3ms-Suspend Counter
      UIR2_EOPFR = 1;           // reset EOP Intr Flag
      continue;                 // try other events
    }

    if (UIR1_RSTF) {
      // USB Reset Signal State detected
      resetUsb();               // call bus-reset handler
      continue;                 // try other events
    }
    break;
  }
}

//----------------------------------------------------------------------------
//  Initiation of IDLE state.
//  Initialize the EP0 transceiver.
//----------------------------------------------------------------------------
void idleInit(void) {
  // initialize endpoint #0
  UCR0  = 0b00010000;
  //        ||||++++-- TP0SIZ=0     reset packet size register
  //        |||+------ RX0E=1       enable EP0 receiver
  //        ||+------- TX0E=0       disable EP0 transmitter
  //        |+-------- Reserved (0)
  //        +--------- T0SEQ=0      reset TX sequence bit
  flag.ep0InComplete = 0;       // Clear complete flag
  flag.ep0OutReady = 0;         // Clear EP0-OUT ready flag
  flag.setupReady = 0;          // Clear SETUP ready flag
  controlState = ST_IDLE;       // goto IDLE state.
}

//----------------------------------------------------------------------------
//  Execution of IDLE state.
//  Check if a SETUP packet received and call parser.
//----------------------------------------------------------------------------
void idleExec(void) {
  if (flag.setupReady) {        // Check if a SETUP packet arrived.
    flag.setupReady = 0;        // Acknowledge
    switch (parseSetup()) {     // parse the SETUP packet
      case ERR_NONE:
        break;
      default:
        forceStall();
        idleInit();             // Restart from IDLE state.
    }
  }
}

//----------------------------------------------------------------------------
//  Initiation of NODATA_STATUS state.
//  STATUS stage of NODATA control transfer is processed.
//----------------------------------------------------------------------------
void nodataStatusInit(byte (*proc)(void)) {
  nodataStatusProc = proc;      // Specify callback invoked after STATUS stage.
  UCR0_T0SEQ = 0;               // Initialize sequence bit
  ep0InStart(0, 0, TRUE);       // Prepare a ZLP as STATUS.
  controlState = ST_NODATA_STATUS;  // state transition.
}

//----------------------------------------------------------------------------
//  Execution of NODATA_STATUS state.
//  Waiting for STATUS stage completion.
//----------------------------------------------------------------------------
void nodataStatusExec(void) {
  if (flag.ep0InComplete) {
    flag.ep0InComplete = 0;     // Acknowledge
    switch ((*nodataStatusProc)()) {    // Actual execution
      case ERR_NONE:
        break;
      default:
        forceStall();
    }
    idleInit();                 // goto IDLE state.
  }
}

void readDataInit(const byte *buf, word len) {
  byte zlp;
  if (len >= setupBuffer.wLength) {
    // Not all data requested.
    len = setupBuffer.wLength;
    zlp = 0;
  } else {
    // All data requested.
    zlp = ((len % EP0_MAXPACKETSIZE) == 0);
  }
  UCR0_T0SEQ = 0;               // Initialize sequence bit
  ep0InStart(buf, len, zlp);
  controlState = ST_READ_DATA;
}

void readDataExec(void) {
  if (flag.ep0InComplete) {     // all data sent
    flag.ep0InComplete = 0;     // Acknowledge
    readStatusInit();           // goto READ_STATUS state.
  }
}

void readStatusInit(void) {
  controlState = ST_READ_STATUS;
}

void readStatusExec(void) {
  if (flag.ep0OutReady) {       // STATUS sent
    flag.ep0OutReady = 0;       // Acknowledge
    idleInit();                 // goto IDLE state.
  }
}

void writeDataInit(byte (*dataProc)(void), void (*statusProc)(void)) {
  writeDataProc = dataProc;
  writeStatusProc = statusProc;
  controlState = ST_WRITE_DATA;
}

void writeDataExec(void) {
  if (flag.ep0OutReady) {
    flag.ep0OutReady = 0;       // Acknowledge
    if ((*writeDataProc)()) {   // receive a packet.
      writeStatusInit();        // goto WRITE_STATUS state.
    }
  }
}

void writeStatusInit(void) {
  ep0InStart(0, 0, TRUE);
  controlState = ST_WRITE_STATUS;
}

void writeStatusExec(void) {
  if (flag.ep0InComplete) {     // Status detected.
    (*writeStatusProc)();       // actual execution.
    idleInit();                 // goto IDLE state.
  }
}

//----------------------------------------------------------------------------
//  Dispatcher task.
//----------------------------------------------------------------------------
void dispatch(void) {
  switch (controlState) {
    case ST_IDLE:
      idleExec();
      break;
    case ST_NODATA_STATUS:
      nodataStatusExec();
      break;
    case ST_READ_DATA:
      readDataExec();
      break;
    case ST_READ_STATUS:
      readStatusExec();
      break;
    case ST_WRITE_DATA:
      writeDataExec();
      break;
    case ST_WRITE_STATUS:
      writeStatusExec();
      break;
    default:
      idleInit();
  }
}

